import datetime
import logging
import os
import traceback
from typing import Any

from elasticsearch import Elasticsearch


def format_time(record: logging.LogRecord, datefmt: str = None) -> str:
    """
    Format the time for the log record.
    :param record: The log record.
    :param datefmt: Optional date format string.
    :return: Formatted timestamp string.
    """
    dt = datetime.datetime.fromtimestamp(record.created, datetime.timezone.utc)
    return dt.strftime(datefmt or "%Y-%m-%dT%H:%M:%SZ")


class LogHandler(logging.Handler):
    def __init__(self, hosts: str, api_key: str, index_name: str, application_name: str, cert_path: str):
        """
        Custom logging handler to send logs to Elasticsearch.
        :param hosts: Comma-separated list of Elasticsearch hosts.
        :param api_key: API key for Elasticsearch authentication.
        :param index_name: Name of the index to store logs.
        :param application_name: Name of the application generating logs.
        :param cert_path: Path to the CA certificate for SSL verification.
        """
        super().__init__()
        self.client = Elasticsearch(
            hosts=[host.strip() for host in hosts.split(",")],
            api_key=api_key,
            verify_certs=False,
            ca_certs=cert_path
        )
        self.index_name = index_name
        self.application_name = application_name

        # Ensure the index exists
        if not self.client.indices.exists(index=self.index_name):
            self.client.indices.create(
                index=self.index_name,
                body={
                    "settings": {
                        "number_of_shards": 1,
                        "number_of_replicas": 1
                    },
                    "mappings": {
                        "properties": {
                            "timestamp": {"type": "date"},
                            "level": {"type": "keyword"},
                            "message": {"type": "text"},
                            "logger": {"type": "keyword"},
                            "filename": {"type": "keyword"},
                            "line": {"type": "integer"},
                            "function": {"type": "keyword"},
                            "application_name": {"type": "keyword"},
                            "stack_trace": {"type": "text"}
                        }
                    }
                }
            )

    def emit(self, record: Any):
        """
        Emit a log record to Elasticsearch.
        :param record: The log record.
        """
        stack_trace = None
        if record.exc_info:
            stack_trace = ''.join(traceback.format_exception(*record.exc_info))

        log_entry = {
            "timestamp": format_time(record),
            "level": record.levelname,
            "message": record.getMessage(),
            "logger": record.name,
            "filename": record.pathname,
            "line": record.lineno,
            "function": record.funcName,
            "application_name": self.application_name,
            "stack_trace": stack_trace,
        }

        try:
            self.client.index(index=self.index_name, body=log_entry)
        except Exception as e:
            print(f"Failed to send log to Elasticsearch: {e}")


def get_logger(name: str = "Logger", level: int = logging.INFO) -> logging.Logger:
    """
    Create and return a logger with Elasticsearch integration using environment variables.
    :param name: Name of the logger.
    :param level: Logging level.
    :return: Configured logger.
    """
    # Retrieve configuration from environment variables
    hosts = os.getenv("ELASTICSEARCH_HOSTS")
    api_key = os.getenv("ELASTICSEARCH_API_KEY")
    index_name = os.getenv("ELASTICSEARCH_INDEX")
    application_name = os.getenv("APPLICATION_NAME")
    cert_path = os.getenv("ELASTICSEARCH_CERT_PATH", "./logger/elasticsearch-ca.crt")
    
    if not all([hosts, api_key, index_name, application_name, cert_path]):
        raise ValueError("One or more required environment variables (ELASTICSEARCH_*) are not set.")

    # Create logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Avoid duplicate handlers if the logger is already configured
    if not logger.handlers:
        # Add Elasticsearch log handler
        es_handler = LogHandler(hosts, api_key, index_name, application_name, cert_path)
        logger.addHandler(es_handler)

        # Optional: Add a console handler for debugging
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
        logger.addHandler(console_handler)

    return logger
